#!/usr/bin/env python
# -*- coding: utf-8 -*-
from threading import Event, Thread


class TimedFunctionCaller(object):
    """!
    Mit dem TimedFunctionCaller können mehrere Funktionen zyklisch aufgerufen werden.
    Die Zykluszeit wird mit interval() festgelegt.
    Funktionen, die in diesem Intervall aufgerufen werden sollen, werden mit register_function()
    festgelegt.
    """
    def __init__(self):
        """!
        Setzt das initiale Intervall auf 3s
        Leert die Funktionsliste
        Startet den Thread
        """
        self._stopped = Event()
        self._interval = 3  # seconds
        self._func_list = []
        self._thread = None

        # add and start the recorder thread
        self.start_calling()

    def __del__(self):
        """
        Destruktor beendet den Thread.
        """
        self._stopped.set()

    def setInterval(self, value):
        """!
        Setzt eine neue Zykluszeit.
        @param value: Zykluszeit in s
        """
        self._interval = value

    def start_calling(self):
        """!
        Legt den DataRecorder Thread an und startet diesen.
        """
        self._stopped.clear()
        self._thread = Thread(target=self._record, name='DataRecorder')
        self._thread.setDaemon(True)
        self._thread.start()

    def stop_calling(self):
        """!
        Stoppt die Ausführung des DataRecorder Threads.
        """
        self._stopped.set()

    def register_function(self, func):
        """!
        Fügt eine Funktion in die Funktionsliste ein.
        Die Funktionen in der Liste werden nacheinander ausgeführt.
        @param func: Funktionsreferenz
        """
        self._func_list.append(func)

    def _record(self):
        """!
        Wartet die Zykluszeit ab und führt dann alle Funktionen in der Funktionsliste
        nacheinander aus.
        """
        while not self._stopped.wait(self._interval):  # the first call is in `interval` secs
            for f in self._func_list:
                f()
