#!/usr/bin/env python
# -*- coding: utf-8 -*-

import Queue
# --- system ---
import logging
from threading import Event, Thread
import time
import serial

from platform import system
if system() == 'Windows':
    import socket

STX = '\x02'        # Start of TeXt - Start einer Botschaft
ETX = '\x03'        # End of TeXt - Ende einer Botschaft

_logger = logging.getLogger(__name__)
ENABLE_DEBUG_LOG = True
ENABLE_RX_DEBUG_LOG = False
ENABLE_TX_DEBUG_LOG = False


class SerialRxTx(object):
    """!
    Diese Klasse ist für die serielle Kommunikation mit der Bedieneinheit verantwortlich.
    """
    def __init__(self, port, baudrate=115200):
        """!
        Initialisiert die Serielle Kommunikation.
        @param port: Name der zu verwendenden Schnittstelle.
        @param baudrate: Übertragungsrate (Default = 115200).
        """
        _logger.debug("SerialRxTx init")
        self._name = "SerialRxTx"

        if system() == 'Windows':
            self.hostname = socket.gethostname()
            self.port = 4444
            self.sock = None
            self._thread_serialrxtx = Thread(target=self._process_win, name=self._name)
        else:
            self._ser = serial.Serial(port, baudrate)
            self._thread_serialrx = Thread(target=self._rx_process, name=self._name)
            self._thread_serialtx = Thread(target=self._tx_process, name=self._name)

        self._checksumchecked = True
        self._firstMessageSent = False
        self._txString = ''
        self._repeatCounter = 0
        self._sendingFileBlocks = False

        # --- buffer ---
        self._input_buffer = Queue.Queue()
        self._output_buffer = Queue.Queue()

        # --- thread ---
        self._rx_process_time = 0.05
        self._rx_stopped = Event()
        self._rx_stopped.clear()
        self._thread_serialrx.setDaemon(True)
        self._thread_serialrx.start()
        self._tx_process_time = 0.05
        self._tx_stopped = Event()
        self._tx_stopped.clear()
        self._thread_serialtx.setDaemon(True)
        self._thread_serialtx.start()

        if ENABLE_DEBUG_LOG:
            _logger.debug("SerialRxTx start")

    def exit(self):
        """!
        Schließt die Schnittstelle und stoppt den Thread.
        """
        self._rx_stopped.set()
        self._tx_stopped.set()
        self._ser.close()

    def write(self, value):
        """!
        Nimmt die zu sendende Botschaft entgegen.
        @param value: Botschaft die versendet werden soll.
        """
        self._input_buffer.put(value)

    def read(self):
        """!
        Funktion zum Lesen der empfangenen Botschaften.
        @return None, wenn nichts empfangen wurde, sonst die empfangene Botschaft.
        """
        if not self._output_buffer.empty():
            rc = self._output_buffer.get()
            self._output_buffer.task_done()
            return rc
        else:
            return None

    def _checkCRC(self, rxString):
        """!
        Funktion zum pruefen der empfangenen Botschaften anhand deren Checksumme
        @return False, wenn Checksumme nicht korrekt, sonst True
        """
        commandlist = rxString.split('$$')[:-1]
        checksumReceived = int(rxString.split('$$')[-1:][0])
        checksumCalculated = 0
        for command in commandlist:
            for c in command:
                checksumCalculated += ord(c)
        checksumCalculated &= 0xFF
        #_logger.debug("calc: {}; rec: {}".format(checksumCalculated, checksumReceived))
        return (checksumCalculated + checksumReceived) == 0

    def set_sending_file_blocks(self, val):
        self._sendingFileBlocks = val
        return True

    def _tx_process(self):
        """!
        Thread zum Senden und Empfangen über die serielle Schnittstelle.
        """

        while not self._tx_stopped.wait(self._tx_process_time):
            if self._sendingFileBlocks:
                if not self._input_buffer.empty():
                    try:
                        tx_msg = STX + self._input_buffer.get() + ETX
                        self._ser.write(tx_msg)
                    except:
                        _logger.warning("tried to send command while in file transfer!")
            else:
                # Senden
                if not self._input_buffer.empty() and self._checksumchecked:                      # Nachricht zum Versenden vorhanden?
                #if not self._input_buffer.empty():
                    self._txString = ''
                    self._checksumchecked = False
                    if self._firstMessageSent:
                        self._input_buffer.task_done()  # Abgeholte Nachricht bestätigen
                    checksum = 0
                    buffer = self._input_buffer.get()
                    for command in buffer.split('$$'):
                        for c in command:
                            checksum += ord(c)
                    checksum &= 0xFF
                    checksum = (~checksum + 1)
                    self._txString = STX + buffer + '$$' + str(checksum) + ETX     # Nachricht holen und Steuerzeichen anhängen

                if not self._checksumchecked:
                    self._ser.write(self._txString.encode())                  # Nachricht senden
                    self._ser.flush()                                   # Warten bis gesendet
                    if ENABLE_TX_DEBUG_LOG:
                        _logger.debug("send: {}".format(self._txString))
                    if not self._firstMessageSent:
                        self._firstMessageSent = True

    def _rx_process(self):
        """!
        Thread zum Senden und Empfangen über die serielle Schnittstelle.
        """
        rxMsgStarted = False
        rxString = ''

        while not self._rx_stopped.wait(self._rx_process_time):
            # Senden
            #if not self._input_buffer.empty() and self._checksumchecked:  # Nachricht zum Versenden vorhanden?
            #    self._txString = ''
            #    self._checksumchecked = False
            #    if self._firstMessageSent:
            #        self._input_buffer.task_done()  # Abgeholte Nachricht bestätigen
            #    checksum = 0
            #    buffer = self._input_buffer.get()
            #    for command in buffer.split('$$'):
            #        for c in command:
            #            checksum += ord(c)
            #    checksum &= 0xFF
            #    checksum = (~checksum + 1)
            #    self._txString = STX + buffer + '$$' + str(checksum) + ETX  # Nachricht holen und Steuerzeichen anhängen
            #
            #if not self._checksumchecked:
            #    self._ser.write(self._txString.encode())  # Nachricht senden
            #    self._ser.flush()  # Warten bis gesendet
            #    if ENABLE_TX_DEBUG_LOG:
            #        _logger.debug("send: {}".format(self._txString))
            #    if not self._firstMessageSent:
            #        self._firstMessageSent = True

            # Empfangen
            while self._ser.inWaiting() > 0:                        # Solange Zeichen ankommen (geht auch if?)
                c = self._ser.read()                                # Zeichen lesen

                if c == STX:                                        # Start der Botschaft
                    rxMsgStarted = True                             # merken
                    rxString = ''                                   # Zeichenkette vorbereiten
                elif c == ETX:                                      # Ende der Botschaft
                    if rxString.find('nack') != -1:
                        self._repeatCounter += 1
                        self._checksumchecked = False               # wurde ein nack empfangen, darf keine weitere Nachricht geschickt werden
                    elif rxString.find('ack') != -1:
                        self._repeatCounter = 0
                        self._checksumchecked = True                # wurde ein ack empfangen, duerfen wieder Nachrichten geschickt werden
                    rxMsgStarted = False                            # merken
                    if self._checkCRC(rxString):                    # empfangene Checksumme mit der eigens berechneten vergleichen
                        if self._repeatCounter > 4:                 # wenn die Nachricht mehr als 5 mal nicht erkannt wurde, mit naechster Nachricht weitermachen
                            self._checksumchecked = True
                        if ENABLE_RX_DEBUG_LOG:
                            _logger.debug("rec: {}".format(rxString))
                        self._output_buffer.put(rxString)           # In Ausgang legen
                    else:
                        _logger.debug("checksum not correct {}".format(rxString))
                elif rxMsgStarted:                                  # Wenn Botschaft gestartet
                    rxString += c                                   # Botschaft Zeichenweise zusammenbauen

    def _process_win(self):
        while not self._rx_stopped.wait(self._rx_process_time):

            if not self._input_buffer.empty():                      # Nachricht zum Versenden vorhanden?
                txString = STX + self._input_buffer.get() + ETX     # Nachricht holen und Steuerzeichen anhängen
                self._input_buffer.task_done()                      # Abgeholte Nachricht bestätigen

                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                try:
                    sock.connect((self.hostname, self.port))
                except socket.error as errMsg:
                    _logger.debug('at {} transmit {} error {}'.format(time.time(), txString, errMsg))
                    return None
                sock.send(str(txString))  # ensure a string is send
                if ENABLE_TX_DEBUG_LOG:
                    _logger.debug("send: {}".format(txString))

                rxString = sock.recv(1024)
                rxString = rxString.replace(STX, '')
                rxString = rxString.replace(ETX, '')
                sock.shutdown(socket.SHUT_RDWR)
                sock.close()
                time.sleep(0.005)    # needed else Windows is not able to 'recycle' the socket addresses in time for re-use
                if ENABLE_RX_DEBUG_LOG:
                    _logger.debug("rec: {}".format(rxString))
                self._output_buffer.put(rxString)               # In Ausgang legen


if __name__ == "__main__":

    logging.basicConfig(level=logging.DEBUG, format='%(asctime)s %(name)-12s %(levelname)-8s %(message)s')

    serRxTx = SerialRxTx('/dev/ttyO1')

    counter = 0
    now = time.time()

    while True:
        rx = serRxTx.read()
        if rx:
            print( rx )

        if time.time() - now > 1:
            now = time.time()
            tx = 'time=={:04d}$$ack'.format(counter)
            serRxTx.write(tx)

            counter += 1
